(* Copyright (C) 2005 Pawel Findeisen, Mikolaj Konarski
 *
 * This file is part of the Dule compiler.
 * The Dule compiler is released under the GNU General Public License (GPL).
 * Please see the file Dule-LICENSE for license information.
 *
 * $Id: test_dags.ml,v 1.1 2005/05/18 23:55:32 mikon Exp $
 *) 

(* The following example shows that type reconstruction
   is polynomial for (some cases of) values with types of exponential size
   but linear representations. *)

let nn = int_of_string Sys.argv.(1) in
let rec declarations n acc = 
    if n=0 then
	acc
    else
	declarations
	    (n-1) ("  id"::(string_of_int n)::" = fun ~x -> x\n"::acc)
in

let rec header n acc = 
    if n=0 then
	"\n  f = fun"::acc
    else
	header (n-1) (" ~id"::(string_of_int n)::acc)
in

let rec condition n acc =
    if n=0 then
	"id"::(string_of_int nn)::acc
    else
	"("::(condition (n-1) (" ~x:id"::(string_of_int (nn-n))::")"::acc))
in

let rec application n acc =
    if n=0 then
	" f"::acc
    else
	application (n-1) (" ~id"::(string_of_int n)::acc)
in

let application = application nn (" ~x\nend"::[]) in
let condition = condition (nn-1) (" ~x "::"\nin"::application) in
let header = header nn (" ~x ->\n    "::condition) in
let declarations = "M = :: sig value v : {} end struct value v = (fun ~t -> {}) ~t: fun ~x -> \nlet\n" :: declarations nn header in
List.iter print_string declarations;;

(* Below are some simpler variants with times obtained 
   under various implementations of sub_app. Note that execution
   is still exponential (without memoizing transformations). *)
(* 
let nn = int_of_string Sys.argv.(1) in
let rec declarations n acc = 
    if n=0 then
	acc
    else
	declarations
	    (n-1) ("let id"::(string_of_int n)::" = fun ~x -> x in\n"::acc)
in

let rec header n acc = 
    if n=0 then
	"\nlet f = fun"::acc
    else
	header (n-1) (" ~id"::(string_of_int n)::acc)
in

let rec condition n acc =
    if n=0 then
	"id"::(string_of_int nn)::acc
    else
	"("::(condition (n-1) (" ~x:id"::(string_of_int (nn-n))::")"::acc))
in

let rec application n acc =
    if n=0 then
	" f"::acc
    else
	application (n-1) (" ~id"::(string_of_int n)::acc)
in

let application = application nn (" ~x:`True\n"::[]) in
let condition = condition (nn-1) (" ~x then\n        id"::(string_of_int nn)::"\n    else\n        id"::(string_of_int nn)::"\nin"::application) in
let header = header nn (" ~x ->\n    if "::condition) in
let declarations = declarations nn header in
List.iter print_string declarations;;

(* original pf, core value
== 1 ==

real	0m0.693s
user	0m0.690s
sys	0m0.000s
== 2 ==

real	0m0.703s
user	0m0.680s
sys	0m0.020s
== 3 ==

real	0m0.696s
user	0m0.690s
sys	0m0.010s
== 4 ==

real	0m0.761s
user	0m0.710s
sys	0m0.030s
== 5 ==

real	0m0.772s
user	0m0.730s
sys	0m0.020s
== 6 ==

real	0m0.859s
user	0m0.780s
sys	0m0.000s
== 7 ==

real	0m0.841s
user	0m0.840s
sys	0m0.000s
== 8 ==

real	0m0.946s
user	0m0.930s
sys	0m0.010s
== 9 ==

real	0m1.165s
user	0m1.150s
sys	0m0.020s
== 10 ==

real	0m1.598s
user	0m1.550s
sys	0m0.050s
== 11 ==

real	0m2.517s
user	0m2.500s
sys	0m0.010s
== 12 ==

real	0m4.884s
user	0m4.760s
sys	0m0.130s
== 13 ==

real	0m8.456s
user	0m8.410s
sys	0m0.040s
== 14 ==

real	0m16.894s
user	0m16.790s
sys	0m0.110s
== 15 ==

real	0m35.224s
user	0m34.260s
sys	0m0.330s
== 16 ==

real	1m12.244s
user	1m10.840s
sys	0m0.580s
== 17 ==

real	2m27.855s
user	2m26.570s
sys	0m1.020s
*)
*)
(*
let nn = int_of_string Sys.argv.(1) in
let rec declarations n acc = 
    if n=0 then
	acc
    else
	declarations
	    (n-1) ("let id"::(string_of_int n)::" = fun ~x -> x in\n"::acc)
in

let rec header n acc = 
    if n=0 then
	"\nlet f = fun"::acc
    else
	header (n-1) (" ~id"::(string_of_int n)::acc)
in

let rec condition n acc =
    if n=0 then
	"id"::(string_of_int nn)::acc
    else
	"("::(condition (n-1) (" ~x:id"::(string_of_int (nn-n))::")"::acc))
in

let rec application n acc =
    if n=0 then
	" f"::acc
    else
	application (n-1) (" ~id"::(string_of_int n)::acc)
in

let application = application nn (" ~x:`True\nend"::[]) in
let condition = condition (nn-1) (" ~x then\n        id"::(string_of_int nn)::"\n    else\n        id"::(string_of_int nn)::"\nin"::application) in
let header = header nn (" ~x ->\n    if "::condition) in
let declarations = "M = :: sig value v : {} end struct value v = (fun ~t -> {}) ~t:\n" :: declarations nn header in
List.iter print_string declarations;;

(* 1
== 1 ==

real	0m0.033s
user	0m0.020s
sys	0m0.010s
== 2 ==

real	0m0.038s
user	0m0.030s
sys	0m0.010s
== 3 ==

real	0m0.045s
user	0m0.040s
sys	0m0.000s
== 4 ==

real	0m0.056s
user	0m0.040s
sys	0m0.010s
== 5 ==

real	0m0.081s
user	0m0.070s
sys	0m0.010s
== 6 ==

real	0m0.116s
user	0m0.110s
sys	0m0.010s
== 7 ==

real	0m0.174s
user	0m0.170s
sys	0m0.010s
== 8 ==

real	0m0.295s
user	0m0.280s
sys	0m0.020s
== 9 ==

real	0m0.526s
user	0m0.520s
sys	0m0.010s
== 10 ==

real	0m1.000s
user	0m0.980s
sys	0m0.020s
== 11 ==

real	0m2.009s
user	0m1.950s
sys	0m0.020s
== 12 ==

real	0m4.071s
user	0m3.880s
sys	0m0.120s
== 13 ==

real	0m8.256s
user	0m8.090s
sys	0m0.060s
== 14 ==

real	0m17.137s
user	0m16.860s
sys	0m0.080s
== 15 ==

real	0m35.241s
user	0m34.700s
sys	0m0.260s
== 16 ==

real	1m15.707s
user	1m13.060s
sys	0m0.720s
== 17 ==

real	2m34.666s
user	2m32.100s
sys	0m1.250s
*)
*)
(*
let nn = int_of_string Sys.argv.(1) in
let rec declarations n acc = 
    if n=0 then
	acc
    else
	declarations
	    (n-1) ("let id"::(string_of_int n)::" = fun ~x -> x in\n"::acc)
in

let rec header n acc = 
    if n=0 then
	"\nlet f = fun"::acc
    else
	header (n-1) (" ~id"::(string_of_int n)::acc)
in

let rec condition n acc =
    if n=0 then
	"id"::(string_of_int nn)::acc
    else
	"("::(condition (n-1) (" ~x:id"::(string_of_int (nn-n))::")"::acc))
in

let rec application n acc =
    if n=0 then
	" f"::acc
    else
	application (n-1) (" ~id"::(string_of_int n)::acc)
in

let application = application nn (" ~x\nend"::[]) in
let condition = condition (nn-1) (" ~x then\n        id"::(string_of_int nn)::"\n    else\n        id"::(string_of_int nn)::"\nin"::application) in
let header = header nn (" ~x ->\n    if "::condition) in
let declarations = "M = :: sig value v : {} end struct value v = (fun ~t -> {}) ~t: fun ~x ->\n" :: declarations nn header in
List.iter print_string declarations;;

(* 2
== 1 ==

real	0m0.035s
user	0m0.030s
sys	0m0.000s
== 2 ==

real	0m0.039s
user	0m0.030s
sys	0m0.010s
== 3 ==

real	0m0.048s
user	0m0.030s
sys	0m0.010s
== 4 ==

real	0m0.061s
user	0m0.050s
sys	0m0.010s
== 5 ==

real	0m0.086s
user	0m0.070s
sys	0m0.010s
== 6 ==

real	0m0.125s
user	0m0.120s
sys	0m0.000s
== 7 ==

real	0m0.200s
user	0m0.190s
sys	0m0.010s
== 8 ==

real	0m0.333s
user	0m0.330s
sys	0m0.000s
== 9 ==

real	0m0.611s
user	0m0.600s
sys	0m0.010s
== 10 ==

real	0m1.172s
user	0m1.160s
sys	0m0.010s
== 11 ==

real	0m2.441s
user	0m2.330s
sys	0m0.010s
== 12 ==

real	0m4.701s
user	0m4.670s
sys	0m0.020s
== 13 ==

real	0m9.589s
user	0m9.530s
sys	0m0.050s
== 14 ==

real	0m20.223s
user	0m19.780s
sys	0m0.180s
== 15 ==

real	0m41.402s
user	0m41.170s
sys	0m0.230s
== 16 ==

real	1m30.492s
user	1m26.670s
sys	0m0.600s
== 17 ==

real	3m7.445s
user	3m1.220s
sys	0m1.510s
*)
*)
(*
let nn = int_of_string Sys.argv.(1) in
let rec declarations n acc = 
    if n=0 then
	acc
    else
	declarations
	    (n-1) ("  id"::(string_of_int n)::" = fun ~x -> x\n"::acc)
in

let rec header n acc = 
    if n=0 then
	"\n  f = fun"::acc
    else
	header (n-1) (" ~id"::(string_of_int n)::acc)
in

let rec condition n acc =
    if n=0 then
	"id"::(string_of_int nn)::acc
    else
	"("::(condition (n-1) (" ~x:id"::(string_of_int (nn-n))::")"::acc))
in

let rec application n acc =
    if n=0 then
	" f"::acc
    else
	application (n-1) (" ~id"::(string_of_int n)::acc)
in

let application = application nn (" ~x\nend"::[]) in
let condition = condition (nn-1) (" ~x "::"\nin"::application) in
let header = header nn (" ~x ->\n "::condition) in
let declarations = "M = :: sig value v : {} end struct value v = (fun ~t -> {}) ~t: fun ~x ->\nlet\n" :: declarations nn header in
List.iter print_string declarations;;

(* 3
== 1 ==

real	0m0.028s
user	0m0.020s
sys	0m0.000s
== 2 ==

real	0m0.034s
user	0m0.020s
sys	0m0.020s
== 3 ==

real	0m0.037s
user	0m0.020s
sys	0m0.010s
== 4 ==

real	0m0.042s
user	0m0.040s
sys	0m0.000s
== 5 ==

real	0m0.052s
user	0m0.040s
sys	0m0.010s
== 6 ==

real	0m0.069s
user	0m0.070s
sys	0m0.000s
== 7 ==

real	0m0.098s
user	0m0.080s
sys	0m0.010s
== 8 ==

real	0m0.144s
user	0m0.150s
sys	0m0.000s
== 9 ==

real	0m0.234s
user	0m0.240s
sys	0m0.000s
== 10 ==

real	0m0.409s
user	0m0.410s
sys	0m0.000s
== 11 ==

real	0m0.772s
user	0m0.760s
sys	0m0.010s
== 12 ==

real	0m1.578s
user	0m1.460s
sys	0m0.020s
== 13 ==

real	0m2.982s
user	0m2.970s
sys	0m0.010s
== 14 ==

real	0m6.556s
user	0m6.150s
sys	0m0.030s
== 15 ==

real	0m13.778s
user	0m12.990s
sys	0m0.130s
== 16 ==

real	0m26.664s
user	0m26.300s
sys	0m0.240s
== 17 ==

real	0m55.774s
user	0m55.040s
sys	0m0.270s

with memo:

== 1 ==

real	0m0.028s
user	0m0.010s
sys	0m0.020s
== 2 ==

real	0m0.031s
user	0m0.020s
sys	0m0.010s
== 3 ==

real	0m0.035s
user	0m0.030s
sys	0m0.000s
== 4 ==

real	0m0.038s
user	0m0.030s
sys	0m0.000s
== 5 ==

real	0m0.041s
user	0m0.040s
sys	0m0.000s
== 6 ==

real	0m0.045s
user	0m0.050s
sys	0m0.000s
== 7 ==

real	0m0.049s
user	0m0.050s
sys	0m0.000s
== 8 ==

real	0m0.055s
user	0m0.060s
sys	0m0.000s
== 9 ==

real	0m0.061s
user	0m0.050s
sys	0m0.010s
== 10 ==

real	0m0.073s
user	0m0.060s
sys	0m0.010s
== 11 ==

real	0m0.083s
user	0m0.080s
sys	0m0.000s
== 12 ==

real	0m0.091s
user	0m0.090s
sys	0m0.000s
== 13 ==

real	0m0.099s
user	0m0.100s
sys	0m0.000s
== 14 ==

real	0m0.111s
user	0m0.100s
sys	0m0.010s
== 15 ==

real	0m0.126s
user	0m0.120s
sys	0m0.010s
== 16 ==

real	0m0.136s
user	0m0.120s
sys	0m0.010s
== 17 ==

real	0m0.148s
user	0m0.150s
sys	0m0.000s
== 18 ==

real	0m0.163s
user	0m0.150s
sys	0m0.010s
== 19 ==

real	0m0.178s
user	0m0.170s
sys	0m0.010s
== 20 ==

real	0m0.200s
user	0m0.200s
sys	0m0.000s
== 21 ==

real	0m0.214s
user	0m0.200s
sys	0m0.010s
== 22 ==

real	0m0.233s
user	0m0.230s
sys	0m0.000s
== 23 ==

real	0m0.255s
user	0m0.260s
sys	0m0.000s
== 24 ==

real	0m0.274s
user	0m0.260s
sys	0m0.010s
== 25 ==

real	0m0.311s
user	0m0.280s
sys	0m0.020s
== 26 ==

real	0m0.320s
user	0m0.310s
sys	0m0.010s
== 27 ==

real	0m0.342s
user	0m0.340s
sys	0m0.000s
== 28 ==

real	0m0.366s
user	0m0.360s
sys	0m0.000s
== 29 ==

real	0m0.396s
user	0m0.400s
sys	0m0.000s
== 30 ==

real	0m0.414s
user	0m0.410s
sys	0m0.000s
== 31 ==

real	0m0.445s
user	0m0.440s
sys	0m0.010s
== 32 ==

real	0m0.486s
user	0m0.470s
sys	0m0.010s
== 33 ==

real	0m0.520s
user	0m0.520s
sys	0m0.000s
== 34 ==

real	0m0.539s
user	0m0.510s
sys	0m0.030s
== 35 ==

real	0m0.568s
user	0m0.540s
sys	0m0.030s
== 36 ==

real	0m0.609s
user	0m0.590s
sys	0m0.020s
== 37 ==

real	0m0.650s
user	0m0.620s
sys	0m0.030s
== 38 ==

real	0m0.692s
user	0m0.690s
sys	0m0.000s
== 39 ==

real	0m0.733s
user	0m0.730s
sys	0m0.000s
== 40 ==

real	0m0.777s
user	0m0.780s
sys	0m0.000s
*)
*)
(*
let nn = int_of_string Sys.argv.(1) in
let rec declarations n acc = 
    if n=0 then
	acc
    else
	declarations
	    (n-1) ("let id"::(string_of_int n)::" = fun ~x -> x in\n"::acc)
in

let rec header n acc = 
    if n=0 then
	"\nfun"::acc
    else
	header (n-1) (" ~id"::(string_of_int n)::acc)
in

let rec condition n acc =
    if n=0 then
	"id"::(string_of_int nn)::acc
    else
	"("::(condition (n-1) (" ~x:id"::(string_of_int (nn-n))::")"::acc))
in

let rec application n acc =
    if n=0 then
	" f"::acc
    else
	application (n-1) (" ~id"::(string_of_int n)::acc)
in

let application = ["end"] in
let condition = condition (nn-1) (" ~x\n"::application) in
let header = header nn (" ~x ->\n"::condition) in
let declarations = "M = :: sig value v : {} end struct value v = (fun ~t -> {}) ~t:\n" :: header in
List.iter print_string declarations;;

(* 4
== 1 ==

real	0m0.024s
user	0m0.020s
sys	0m0.000s
== 2 ==

real	0m0.025s
user	0m0.030s
sys	0m0.000s
== 3 ==

real	0m0.026s
user	0m0.020s
sys	0m0.010s
== 4 ==

real	0m0.026s
user	0m0.010s
sys	0m0.010s
== 5 ==

real	0m0.027s
user	0m0.030s
sys	0m0.000s
== 6 ==

real	0m0.029s
user	0m0.020s
sys	0m0.010s
== 7 ==

real	0m0.032s
user	0m0.020s
sys	0m0.000s
== 8 ==

real	0m0.032s
user	0m0.020s
sys	0m0.010s
== 9 ==

real	0m0.034s
user	0m0.010s
sys	0m0.020s
== 10 ==

real	0m0.036s
user	0m0.030s
sys	0m0.010s
== 11 ==

real	0m0.036s
user	0m0.040s
sys	0m0.000s
== 12 ==

real	0m0.038s
user	0m0.040s
sys	0m0.000s
== 13 ==

real	0m0.040s
user	0m0.040s
sys	0m0.000s
== 14 ==

real	0m0.043s
user	0m0.040s
sys	0m0.010s
== 15 ==

real	0m0.044s
user	0m0.050s
sys	0m0.000s
== 16 ==

real	0m0.047s
user	0m0.040s
sys	0m0.000s
== 17 ==

real	0m0.050s
user	0m0.050s
sys	0m0.000s
*)
*)
